/*
 * exe.c - exe archive file handler code
 * (C)Copyright 2000 by TAJIRI Yasuhiro
 * This file is part of Enfle.
 *
 * Last Modified: Sat Jun 10 21:47:52 2000.
 * $Id: exe.c,v 1.1 2000/06/10 12:47:59 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#include "enfle.h"
#include "plugin.h"
#include "dlist.h"
#include "utils.h"
#include "archive.h"
#include "exe.h"

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
archiver_exe_get_plugininfo(PluginInfo *p)
#endif
{
    p->version = 1;
    p->type = _Archiver;
    p->pluginname = "Exe Plugin (SPI) Archiver Plugin Version 0.1.1";
    p->pluginshortname = FORMAT_NAME;
    p->author = "TAJIRI Yasuhiro";
    p->dlhandle = NULL; /* set by plugin_load */
    p->functions.archiver.open = exe_archive_open;

    return 1;
}


static void
exe_free(void *data)
{
    if (data != NULL) {
	free(data);
	data = NULL;
    }
}

void
destroy_exe_info(Archive* ar)
{
    EXE_info *info = ar->info;

    exe_free(info->positions);
    exe_free(info->filesizes);
    exe_free(info->names);
    exe_free(info->data);
}

int
init_exe_info(EXE_info *info, int nfiles)
{
    if ((info->positions =
	    (unsigned long *)calloc(nfiles, sizeof(unsigned long))) == NULL) {
        fprintf(stderr, "init_exe_archive_info: no memory for positions\n");
        return 0;
    }
    if ((info->filesizes =
	     (unsigned long *)calloc(nfiles, sizeof(unsigned long))) == NULL) {
        fprintf(stderr, "init_exe_archive_info: no memory for filesizes\n");
        free(info->positions);
        return 0;
    }
    
    if ((info->names =
	     (unsigned char **)calloc(nfiles, sizeof(unsigned char*))) == NULL) {
        fprintf(stderr, "init_exe_archive_info: no memory for filesizes\n");
        free(info->filesizes);
        free(info->positions);
        return 0;
    }
    return 1;
}

int
exe_archive_open(Archive *ar)
{
    EXE_info *info;
    int i;
    Dlist* list;
    Dlist_data* data;
    int num;

#ifdef DEBUG
    fprintf(stderr, "exe_archive_open: %s open \n", ar->filename);
#endif
    if ((ar->fp = fopen(ar->filename, "rb")) == NULL) {
        fprintf(stderr, "exe_archive_open: %s open failed\n", ar->filename);
        return 0;
    }
    
    if ((ar->info = (EXE_info *)calloc(1, sizeof(EXE_info))) == NULL) {
        fprintf(stderr, "exe_archive_open: no memory for info");
        return 0;
    }
    info = ar->info;
    info->data = NULL;

    if((list = getInfoList(ar->fp))==NULL){
        free(ar->info);
	ar->info = NULL;
        fclose(ar->fp);
        return 0;
        
    }
    
    num = dlist_getndata(list);
    
    ar->nfiles=num;
    init_exe_info(info, num);
    data=dlist_gettop(list);
    for(i=0;i<num;i++){
        EXE_RES_INFO* list_info = (EXE_RES_INFO*)data->data;
        info->positions[i] = list_info->pos;
        info->filesizes[i] = list_info->size;
        info->names[i]=list_info->name;
        data=data->next;

    }
    ar->format = FORMAT_NAME;
    ar->select = exe_archive_select;
    ar->seek = exe_archive_seek;
    ar->tell = exe_archive_tell;
    ar->read = exe_archive_read;
    ar->close = exe_archive_close;

    return 1;
}

int
exe_archive_select(Archive *ar, int index)
{
    EXE_info *info = ar->info;
    if (info->data != NULL) {
        free(info->data);
        info->data = NULL;
    }
    info->position = info->positions[index];
    fseek(ar->fp,info->position,SEEK_SET);
    ar->size = info->filesizes[index];
    if ((info->data = (unsigned char *)malloc(ar->size)) == NULL) {
        fprintf(stderr, "exe_archive_select: no memory for data\n");
        return 0;
    }
    fread(info->data,ar->size,1,ar->fp);
    ar->offset = info->position;
    fseek(ar->fp, ar->offset, SEEK_SET);
    
    strncpy(ar->ifname,info->names[index],8);
    ar->ifname[8]='\0';
    

#ifdef DEBUG
    printf("exe_archive_select:\n");
    printf("    ar->ifname: %s (%d)\n", ar->ifname, index);
    printf("    ar->offset: %d\n", ar->offset);
    printf("    ar->size: %d\n", ar->size);
#endif

    return 1;
}

int
exe_archive_seek(Archive *ar, long pos, int whence)
{
    EXE_info *info = ar->info;

    switch (whence) {
        case SEEK_SET:
	    if (pos < 0 || pos > ar->size)
		return -1;
	    info->position = pos;
	    break;
        case SEEK_CUR:
	    if (info->position + pos < 0 || info->position + pos > ar->size)
		return -1;
	    info->position += pos;
	    break;
        case SEEK_END:
	    if (ar->size + pos < 0 || ar->size + pos > ar->size)
		return -1;
	    info->position = ar->size + pos;
    }

    return 0;
}

int
exe_archive_tell(Archive *ar)
{
    EXE_info *info = ar->info;

    return info->position;
}

int
exe_archive_read(Archive *ar, unsigned char *buf, int size)
{
    EXE_info *info = ar->info;
    int rsize = size;

    if (rsize < 0)
	return 0;
    if (info->position + rsize > ar->size)
        rsize = ar->size - info->position;
    memcpy(buf, info->data + info->position, rsize);
    info->position += rsize;

    return rsize;
}

int
exe_archive_close(Archive *ar)
{
    int f = fclose(ar->fp);

    destroy_exe_info(ar);
    free(ar->info);

    return f;
}
