/*
 * misc.c -- misc archive handler
 * Copyright (C) 1999 TAJIRI Yasuhiro
 * This file is part of Enfle.
 *
 * Last Modified: Sun Jan  9 14:44:26 2000.
 * $Id: misc.c,v 1.7 2000/01/23 11:18:59 sian Exp $
 *
 * by courtesy of Linux kernel source 
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "enfle.h"
#include "plugin.h"
#include "misc.h"
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/wait.h>
#define FORMAT_NAME "MISC"

#define ENTRY_ENABLED 1		
#define	ENTRY_MAGIC 8		/* not filename detection */
static int load_inifile(char *filename);

struct fmt_entry {
	struct fmt_entry *next;
	long id;
	int flags;			/* type, status, etc. */
	int offset;			/* offset of magic */
	int size;			/* size of magic/mask */
	char *magic;			/* magic or filename extension */
	char *mask;			/* mask, NULL for exact match */
	char *interpreter;		/* filename of interpreter */
	char *proc_name;
};
static struct fmt_entry *entries=NULL;
static int free_id = 1;

static char *tmpdir;   /*temporary directory*/

static int
scan_directory(char *dirpath,Archive* ar);



int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
archiver_misc_get_plugininfo(PluginInfo *p)
#endif
{
  
  p->version = 1;
  p->type = _Archiver;
  p->pluginname = "Misc Format Archiver Plugin version 0.3";
  p->pluginshortname = FORMAT_NAME;
  p->author = "TAJIRI Yasuhiro";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.archiver.open = misc_archive_open;
  return 1;
}


/*
 * Unregister one entry
 */
#if 0
static void clear_entry(int id)
{
	struct fmt_entry **ep, *e;

	ep = &entries;
	while (*ep && ((*ep)->id != id))
		ep = &((*ep)->next);
	if ((e = *ep))
		*ep = e->next;
	if (e) {
		free(e);
	}
}
#endif

/*
 * Clear all registered formats
 */
static void clear_entries(void)
{
	struct fmt_entry *e, *n;

	n = entries;
	entries = NULL;
	while ((e = n)) {
		n = e->next;
		free(e);
	}
}

#if 0
/*
 * Find entry through id and lock it
 */
static struct fmt_entry *get_entry(int id)
{
	struct fmt_entry *e;

	e = entries;
	while (e && (e->id != id))
		e = e->next;
	return e;
}
#endif


/* 
 * Check if we support the binfmt
 * if we do, return the binfmt_entry, else NULL
 * locking is done in load_misc_binary
 */
static struct fmt_entry *check_file(Archive* ar)
{
  struct fmt_entry *e;
  char *p = strrchr(ar->filename, '.');
  int j;

  e = entries;
  while (e) {
    if (e->flags & ENTRY_ENABLED) {
      if (!(e->flags & ENTRY_MAGIC)) {
	if (p && !strcmp(e->magic, p + 1))
	  return e;
      } else {
	char* buf;
	if((buf = malloc(sizeof(char) * e->size)) == NULL)
	  return NULL;
	j = 0;

	fseek(ar->fp, e->offset, SEEK_SET);
	if(fread(buf, sizeof(char), e->size, ar->fp) != e->size) {
	  free(buf);
	  return NULL;
	}

	while ((j < e->size) &&
	       !((buf[j] ^ e->magic[j])
		 & (e->mask ? e->mask[j] : 0xff)))
	  j++;
	free(buf);
	if (j == e->size)
	  return e;
      }
    }
    e = e->next;
  }
  return NULL;
}

/*
 * parses and copies one argument enclosed in del from *sp to *dp,
 * recognising the \x special.
 * returns pointer to the copied argument or NULL in case of an
 * error (and sets err) or null argument length.
 */
static char *copyarg(char **dp, const char **sp, int *count,
		     char del, int special, int *err)
{
	char *res = *dp;
    int c = 0;
    
	while (!*err && ((c = *((*sp)++)), (*count)--) && (c != del)) {
		switch (c) {
		case '\\':
			if (special && (**sp == 'x')) {
                c = toupper((int)*(++*sp));
				if (!isxdigit(c))
					*err = -EINVAL;
				**dp = (c - (isdigit(c) ? '0' : 'A' - 10)) * 16;
                c = toupper(*(++*sp));
				if (!isxdigit(c))
					*err = -EINVAL;
				*((*dp)++) += c - (isdigit(c) ? '0' : 'A' - 10);
				++*sp;
				*count -= 3;
				break;
			}
		default:
			*((*dp)++) = c;
		}
	}
	if (*err || (c != del) || (res == *dp))
		res = NULL;
	else if (!special)
		*((*dp)++) = '\0';
	return res;
}

/*
 * This registers a new binary format, it recognises the syntax
 * ':name:type:offset:magic:mask:interpreter:'
 * where the ':' is the IFS, that can be chosen with the first char
 */
static int registerfmt(const char *buffer,unsigned long count)
{
	const char *sp;
	char del, *dp;
	struct fmt_entry *e;
	int memsize, cnt = count - 1, err;

	/* some sanity checks */
	err = -EINVAL;
	if ((count < 11) || (count > 256))
		goto _err;

	err = -ENOMEM;
	memsize = sizeof(struct fmt_entry) + count;
	if (!(e = (struct fmt_entry *) malloc(memsize)))
		goto _err;

	err = 0;
	sp = buffer + 1;
	del = buffer[0];
	dp = (char *)e + sizeof(struct fmt_entry);

	e->proc_name = copyarg(&dp, &sp, &cnt, del, 0, &err);

	/* we can use bit 3 of type for ext/magic
	   flag due to the nice encoding of E and M */
	if ((*sp & ~('E' | 'M')) || (sp[1] != del))
		err = -EINVAL;
	else
		e->flags = (*sp++ & (ENTRY_MAGIC | ENTRY_ENABLED));
	cnt -= 2; sp++;

	e->offset = 0;
	while (cnt-- && isdigit(*sp))
		e->offset = e->offset * 10 + *sp++ - '0';
	if (*sp++ != del)
		err = -EINVAL;

	e->magic = copyarg(&dp, &sp, &cnt, del, (e->flags & ENTRY_MAGIC), &err);
	e->size = dp - e->magic;
	e->mask = copyarg(&dp, &sp, &cnt, del, 1, &err);
	if (e->mask && ((dp - e->mask) != e->size))
		err = -EINVAL;
	e->interpreter = copyarg(&dp, &sp, &cnt, del, 0, &err);
	e->id = free_id++;

	/* more sanity checks */
	if (err || !(!cnt || (!(--cnt) && (*sp == '\n'))) ||
	    (e->size < 1) || ((e->size + e->offset) > 127) ||
	    !(e->proc_name) || !(e->interpreter) )
		goto free_err;

	e->next = entries;
	entries = e;
	err = count;
_err:
	return err;
free_err:
	free(e);
	err = -EINVAL;
	goto _err;
}

static int load_inifile(char *filename)
{
    FILE * fp;
    char buf[257];

    if ((fp = fopen(filename,"r")) == NULL)
      return 1;
    buf[256]=0;
    while(fgets(buf,256,fp)!=NULL){
        int count = strlen(buf);
        if(count != registerfmt(buf,count))
        {
            fprintf(stderr,"init file format error\n");
            fclose(fp);
            return 1;
        }
    }
    fclose(fp);
    return 0;
}

int
misc_archive_open(Archive *ar)
{
    struct fmt_entry *e;
    char tmpsh[256];
    static char *preName = NULL;
    /* char curTmpDir[500]; */
    FILE* fp;
    pid_t pid_child;
    MISC_info *info = NULL;
    char rcfile[50];
    char *home;

#ifdef DEBUG
  fprintf(stderr, "misc_archive_open(): open %s\n", ar->filename);
#endif

    if (ar->format != NULL)
      return 0;
    if ((ar->fp = fopen(ar->filename, "rb")) == NULL)
        return 0;

    fseek(ar->fp, 0L, SEEK_END);
    ar->asize = ftell(ar->fp);
    fseek(ar->fp, 0L, SEEK_SET);
    ar->nfiles=0;
    home = getenv("HOME");
    if (home == NULL)
        goto error;
    if (entries == NULL) {
        sprintf(rcfile,"%s/%s",home,".miscloader");
        if(load_inifile(rcfile))
            goto error;
    }
    if ((info = calloc(1, sizeof(MISC_info))) == NULL) {
        fprintf(stderr, "No enough memory for info\n");
        archive_close(ar);
        goto clear_error;
    }
    info->fname=NULL;
    e = check_file(ar);
    if(e==NULL)
        goto clear_error;
    sprintf(tmpsh,"/tmp/.%s%d",getenv("USER"),(int)(getpid()));
    /* sprintf(tmpsh,"/tmp/.%s%d%s",getenv("USER"),(int)(getpid()),ar->filename); */
    tmpdir=strdup(tmpsh);
    mkdir(tmpdir,0755);
    sprintf(tmpsh,"%s/expand.sh",tmpdir);
    fp=fopen(tmpsh,"w");
    fprintf(fp,e->interpreter,tmpdir,ar->filename);
    fclose(fp);
    if (preName == NULL || (strcmp(ar->filename,preName) != 0)){
       if (preName)
	 free(preName);
       preName=strdup(ar->filename);
       pid_child = fork();
        if (pid_child == 0) {
            /* child process */
            pid_t mine = getpid();
            close(1);
            setpgid(mine, mine);
            execl("/bin/sh","/bin/sh",tmpsh,NULL);
            _exit(-1);
        } else if (pid_child < 0) {
            fprintf(stderr, "fork failed");
            exit(0);
            goto clear_error;
        } else {
            int status;
            waitpid(pid_child,&status,0);
        }
    }
    scan_directory(tmpdir,ar);
    if (ar->nfiles) {
        ar->info = info;
        ar->format = FORMAT_NAME;
        ar->select = misc_archive_select;
        ar->close  = misc_archive_close;
	return 1;
    }
 clear_error:
    clear_entries();
 error:
    if(info != NULL){
        int i;
        for (i = 0; i < ar->nfiles; i++)
            free(info->fname[i]);
        if (info->fname)
	  free(info->fname);
        free(info);
    }
    fclose(ar->fp);
    return 0;
}


static int
scan_directory(char *dirpath,Archive* ar)
{
    char *filename;
    DIR *dir;
    struct dirent *ent;
    struct stat statbuf;
    MISC_info *info=ar->info;
    int count = 0;
    if ((dir = opendir(dirpath)) == NULL) {
        char buf[256];
        sprintf(buf, "scan__directory: opendir: %s", dirpath);
        perror(buf);
        return 0;
    }
    
    while ((ent = readdir(dir))) {
      if (!strcmp(ent->d_name, ".") ||
	  !strcmp(ent->d_name, "..") ||
	  !strcmp(ent->d_name, "expand.sh"))
            continue;
        if ((filename = malloc(strlen(dirpath) + strlen(ent->d_name) + 2)) == NULL) {
            fprintf(stderr, "scan__directory: FATAL: No enough memory for filename\n");
            closedir(dir);
            exit(-1);
        }
        sprintf(filename, "%s/%s", dirpath, ent->d_name);
        if (!stat(filename, &statbuf)) {
            if (S_ISDIR(statbuf.st_mode))
                count += scan_directory(filename,ar);
            if (S_ISREG(statbuf.st_mode)) {
                   if (ar->nfiles % 100 == 0) {
                       int s = ar->nfiles + 101 ;
                       if ((info->fname =
                            realloc(info->fname, sizeof(char*) * s)) == NULL ){
                           fprintf(stderr, "No enough memory for info\n");
			   /* archive_close(info->ar); */
                           fclose(ar->fp);
                           return 0;
                       }
                   }
                   info->fname[ar->nfiles] = strdup(filename);
                   ar->nfiles++;
                   count++;
            }
        }
        free(filename);
    }
    closedir(dir);
    return count;
}

int
misc_archive_select(Archive *ar, int index)
{
  MISC_info *info = ar->info;
  if(ar==NULL)
      return 0;
  if (index < 0)
    return 0;
  fclose(ar->fp);
  ar->fp=fopen(info->fname[index],"r");
  fseek(ar->fp, 0L, SEEK_END);
  ar->size = ftell(ar->fp);
  fseek(ar->fp, 0L, SEEK_SET);

  strncpy(ar->ifname, info->fname[index], 16);
#ifdef DEBUG
  fprintf(stderr, "name = %s\n", info->fname[index]);
#endif
  ar->offset = 0;
  return 1;
}

int
misc_archive_close(Archive *ar)
{
  MISC_info *info = ar->info;
  int i, f = fclose(ar->fp);
  pid_t pid_child;

  clear_entries();
  for (i = 0; i < ar->nfiles; i++)
    free(info->fname[i]);
  free(info->fname);
  free(ar->info);
  ar->info=NULL;
  pid_child = fork();
  if (pid_child == 0) {
      /* child process */
      pid_t mine = getpid();
//      close(1);
      setpgid(mine, mine);
      execl("/bin/rm","/bin/rm","-rf",tmpdir,NULL);
      _exit(-1);
  } else if (pid_child < 0) {
      fprintf(stderr, "fork failed");
      return 0;
  } else {
      int status;
      waitpid(pid_child,&status,0);
  }
  free(tmpdir);
  return f;
}
