/*
 * pak_fan.c -- PAK(FAN) archive handler
 * (C)Copyright 1999 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Mon Sep 20 20:19:05 1999.
 * $Id: pak_fan.c,v 1.6 1999/09/20 13:00:21 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "utils.h"
#include "plugin.h"
#include "pak.h"
#include "pak_fan.h"

int
pak_fan_archive_open(Archive *ar)
{
  char *header, *h;
  PAK_info *info = ar->info;
  int i, j, n;

  /* seek to the top of table */
  fseek(ar->fp, 4, SEEK_SET);

  /* allocate memory, read and calculate key */
  if ((header = malloc(28 * ar->nfiles)) == NULL) {
    fclose(ar->fp);
    free(ar->info);
    return 0;
  }

  if (fread(header, 1, 28 * ar->nfiles, ar->fp) != 28 * ar->nfiles) {
    fclose(ar->fp);
    free(header);
    free(ar->info);
    return 0;
  }

  if (!init_pak_info(info, ar->nfiles)) {
    fclose(ar->fp);
    free(header);
    free(ar->info);
    return 0;
  }

  h = header;
  n = ar->nfiles;
  for (i = 0; i < ar->nfiles;) {
    if (strlen(h) >= 16) {
      fclose(ar->fp);
      free(header);
      free(ar->info);
      return 0;
    }
    if (strcmp(h + strlen(h) - 4, ".grp")) {
      h += 28;
      ar->nfiles--;
      continue;
    }

    strcpy(info->filenames[i], h);
    info->sizes[i] = get_little_dword(h + 16);
    info->flags[i] = get_little_dword(h + 20);
    info->offsets[i] = get_little_dword(h + 24);
    h += 28;
    i++;
  }

  h = header;
  for (j = 0; j < n; j++) {
    if (strcmp(h + strlen(h) - 4, ".c16")) {
      h += 28;
      continue;
    }

    strcpy(info->filenames[i], h);
    info->sizes[i] = get_little_dword(h + 16);
    info->flags[i] = get_little_dword(h + 20);
    info->offsets[i] = get_little_dword(h + 24);
    h += 28;
    i++;
  }
  info->nfiles = i;

  /*
   * ar->nfiles has # of *.grp
   * info->nfiles has # of *.grp + *.c16
   * ar->filenames[0..ar->nfiles - 1] = *.grp
   * ar->filenames[ar->nfiles.. info->nfiles - 1] = *.c16
   */

#ifdef DEBUG
  fprintf(stderr, "PAK(FANDISK) identified.\n");
#endif

  free(header);

  ar->format = FORMAT_NAME;
  ar->select = pak_fan_archive_select;
  ar->seek = pak_fan_archive_seek;
  ar->tell = pak_fan_archive_tell;
  ar->read = pak_fan_archive_read;
  ar->close = pak_archive_close;

  return 1;
}

static int
find_file(Archive *ar, char *filename)
{
  PAK_info *info = ar->info;
  int i;

  for (i = ar->nfiles; i < info->nfiles; i++)
    if (!strcmp(info->filenames[i], filename))
      return i;

  return -1;
}

int
pak_fan_archive_select(Archive *ar, int index)
{
  PAK_info *info = ar->info;
  int s, c16index;
  char *c16name;

  strncpy(ar->ifname, info->filenames[index], 15);
  ar->offset = info->offsets[index];
  ar->size = info->sizes[index] + 2 + 4 * 256 + 4;
#ifdef DEBUG
  fprintf(stderr, "pak_fan_archive_select: %d: %s(%d bytes) offset = %08X\n", index, ar->ifname, ar->size, ar->offset);
#endif

  fseek(ar->fp, ar->offset, SEEK_SET);
  if (info->data != NULL)
    free(info->data);
  if ((info->data = malloc(ar->size)) == NULL) {
    fprintf(stderr, "pak_fan_archive_select: No enough memory\n");
    return 0;
  }
  if ((s = fread(info->data + 0x406, 1, ar->size - 0x406, ar->fp)) == 0)
    return 0;
  if ((c16name = replace_ext(ar->ifname, ".c16")) == NULL) {
    fprintf(stderr, "pak_fan_archive_select: No enough memory\n");
    return 0;
  }
  if ((c16index = find_file(ar, c16name)) < 0) {
    fprintf(stderr, "pak_fan_archive_select: Corresponding c16 file not found\n");
    return 0;
  }

  free(c16name);
  fseek(ar->fp, info->offsets[c16index], SEEK_SET);
  fread(info->data, 1, info->sizes[c16index], ar->fp);
  *(int *)(info->data + 2 + 4 * 256) = info->flags[index];
  info->position = 0;

  return 1;
}

int
pak_fan_archive_seek(Archive *ar, long pos, int whence)
{
  PAK_info *info = ar->info;

  switch (whence) {
  case SEEK_SET:
    if (pos < 0 || pos > ar->size)
      return -1;
    info->position = pos;
    break;
  case SEEK_CUR:
    if (info->position + pos < 0 || info->position + pos > ar->size)
      return -1;
    info->position += pos;
    break;
  case SEEK_END:
    if (ar->size + pos < 0 || ar->size + pos > ar->size)
      return -1;
    info->position = ar->size + pos;
  }

  return 0;
}

int
pak_fan_archive_tell(Archive *ar)
{
  PAK_info *info = ar->info;

  return info->position;
}

int
pak_fan_archive_read(Archive *ar, unsigned char *buf, int size)
{
  PAK_info *info = ar->info;
  int rsize = size;

  if (rsize < 0)
    return 0;
  if (info->position + rsize > ar->size)
    rsize = ar->size - info->position;
  memcpy(buf, info->data + info->position, rsize);
  info->position += rsize;

  return rsize;
}
