/*
 * pak_lvn.c -- PAK(LVN) archive handler
 * (C)Copyright 1999 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Wed Aug 25 00:46:59 1999.
 * $Id: pak_lvn.c,v 1.3 1999/08/24 17:09:22 sian Exp $
 *
 * by courtesy of TF(The author of Leaf CG viewer)
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "plugin.h"
#include "pak.h"
#include "pak_lvn.h"

static void
calculate_key(PAK_info *info, unsigned char *header)
{
  info->key[0] = header[11];
  info->key[1] = (header[12] - 0x0a) & 0xff;
  info->key[2] = header[13];
  info->key[3] = header[14];
  info->key[4] = header[15];
  info->key[5] = (header[38] - header[22] + info->key[0]) & 0xff;
  info->key[6] = (header[39] - header[23] + info->key[1]) & 0xff;
  info->key[7] = (header[62] - header[46] + info->key[2]) & 0xff;
  info->key[8] = (header[63] - header[47] + info->key[3]) & 0xff;
  info->key[9] = (header[20] - header[36] + info->key[3]) & 0xff; 
  info->key[10] = (header[21] - header[37] + info->key[4]) & 0xff;
}

static void
regularize_filename(char *filename)
{
  char str[13];
  int i = 0;

  strcpy(str, filename);
  while (i < 8 && str[i] != ' ')
    i++;
  filename[i++] = '.';
  filename[i++] = str[8];
  filename[i++] = str[9];
  filename[i++] = str[10];
  filename[i]   = '\0';
}

static void
extract_header(PAK_info *info, unsigned char *header, int nfiles)
{
  int i, j, k = 0;
  int buf[4];

  for (i = 0; i < nfiles; i++) {
    for (j = 0; j < 12; j++) {
      info->filenames[i][j] = (*header++ - info->key[k]) & 0xff;
      k = (++k) % PAK_KEY_LEN;
    }
    info->filenames[i][12] = '\0';
    regularize_filename(info->filenames[i]);

    for (j = 0; j < 4; j++) {
      buf[j] = (*header++ - info->key[k]) & 0xff;
      k = (++k) % PAK_KEY_LEN;
    }
    info->offsets[i] = (buf[3] << 24) | (buf[2] << 16) |
                       (buf[1] << 8)  | buf[0];
    for (j = 0; j < 4; j++) {
      buf[j] = (*header++ - info->key[k]) & 0xff;
      k = (++k) % PAK_KEY_LEN;
    }
    info->sizes[i] = (buf[3] << 24) | (buf[2] << 16) |
                     (buf[1] << 8)  | buf[0];
    /* skip next data (the head of the next file) */
    header += 4;
    k = (k + 4) % PAK_KEY_LEN;
  }
}

int
pak_lvn_archive_open(Archive *ar)
{
  char *header;
  PAK_info *info = ar->info;

  /* seek to the top of table */
  fseek(ar->fp, (long)(-24 * ar->nfiles), SEEK_END);

  /* allocate memory, read and calculate key */
  if ((header = malloc(24 * ar->nfiles)) == NULL) {
    fclose(ar->fp);
    free(ar->info);
    return 0;
  }

  if (fread(header, 1, 24 * ar->nfiles, ar->fp) != 24 * ar->nfiles) {
    fclose(ar->fp);
    free(header);
    free(ar->info);
    return 0;
  }

  calculate_key(info, header);
  if (!init_pak_info(info, ar->nfiles)) {
    fclose(ar->fp);
    free(header);
    free(ar->info);
    return 0;
  }
  extract_header(info, header, ar->nfiles);

  free(header);

  ar->format = FORMAT_NAME;
  ar->select = pak_lvn_archive_select;
  ar->seek = NULL;
  ar->tell = NULL;
  ar->read = pak_lvn_archive_read;
  ar->close = pak_archive_close;

  return 1;
}

int
pak_lvn_archive_select(Archive *ar, int index)
{
  PAK_info *info = ar->info;

  strncpy(ar->ifname, info->filenames[index], 15);
  ar->offset = info->offsets[index];
  ar->size = info->sizes[index];
#ifdef DEBUG
  fprintf(stderr, "pak_lvn_archive_select: %d: %s(%d bytes) offset = %08X\n", index, ar->ifname, ar->size, ar->offset);
#endif
  fseek(ar->fp, ar->offset, SEEK_SET);

  return 1;
}

int
pak_lvn_archive_read(Archive *ar, unsigned char *buf, int size)
{
  PAK_info *info = ar->info;
  int k = (ftell(ar->fp) - ar->offset) % PAK_KEY_LEN;
  int i, s;

  if ((s = fread(buf, 1, size, ar->fp)) == 0)
    return 0;

  for (i = 0; i < s; i++) {
    buf[i] = (buf[i] - info->key[k]) & 0xff;
    k = (++k) % PAK_KEY_LEN;
  }

  return s;
}
