/*
 * tar.c -- tar file handler 
 * (C)Copyright 1999 by TAJIRI Yasuhiro
 *   .tar.gz, .tar.bz2 support by Hiroshi Takekawa
 * This is part of Enfle.
 *
 * octal_value & tar_checksum TiMidityή
 *
 * Last Modified: Thu Jun 15 04:01:51 2000.
 * $Id: tar.c,v 1.7 2000/06/14 19:45:20 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <ctype.h>

#include "enfle.h"
#include "plugin.h"
#include "tar.h"

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
archiver_tar_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Archiver;
  p->pluginname = "TAR Format Archiver Plugin version 0.3";
  p->pluginshortname = FORMAT_NAME;
  p->author = "TAJIRI Yasuhiro";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.archiver.open = tar_archive_open;

  return 1;
}

static long
octal_value(char *s, int len)
{
  long val;

  while (len > 0 && !isdigit((int)(unsigned char) * s)) {
    s++;
    len--;
  }

  val = 0;
  while (len > 0 && isdigit((int)(unsigned char) * s)) {
    val = ((val << 3) | (*s - '0'));
    s++;
    len--;
  }
  return val;
}

static int
tar_checksum(char *hdr)
{
  int i;
  long recorded_sum;
  long unsigned_sum;		/* the POSIX one :-) */
  long signed_sum;		/* the Sun one :-( */

  recorded_sum = octal_value(hdr + 148, 8);
  unsigned_sum = 0;
  signed_sum = 0;
  for (i = 0; i < TARBLKSIZ; i++) {
    unsigned_sum += 0xFF & hdr[i];
    signed_sum   += hdr[i];
  }

  /* Adjust checksum to count the "chksum" field as blanks.  */
  for(i = 0; i < 8; i++) {
    unsigned_sum -= 0xFF & hdr[148 + i];
    signed_sum   -= hdr[i];
  }
  unsigned_sum += ' ' * 8;
  signed_sum   += ' ' * 8;

  return unsigned_sum == recorded_sum || signed_sum == recorded_sum;
}

int
tar_archive_open(Archive *ar)
{
  TAR_info *info;
  int size, flen, sizeb;
  char hdr[TARHDRSIZ];
  Archive *ar2;

  if (ar->format && !strcmp(ar->format, FORMAT_NAME))
    return 0;

  if ((ar2 = archive_open_from_plugin(ar->filename, FORMAT_NAME)) == NULL)
    return 0;

#ifdef DEBUG
  fprintf(stderr, "tar_archive_open(): open %s\n", ar->filename);
#endif

  if ((info = calloc(1, sizeof(TAR_info))) == NULL) {
    fprintf(stderr, "No enough memory for info\n");
    archive_close(ar2);
    return 0;
  }
  info->ar = ar2;
  info->ptr = NULL;

  if (archive_seek(info->ar, 0L, SEEK_END) < 0)
    return 0;
  ar->asize = archive_tell(info->ar);
  archive_seek(info->ar, 0L, SEEK_SET);
  ar->nfiles = 0;

  while (1) {
    if (archive_read(info->ar, hdr, TARHDRSIZ) != TARHDRSIZ) {
      archive_close(info->ar);
      free(info);
      return 0;
    }

    if (hdr[0] == '\0')
      break;
    if (!tar_checksum(hdr)){
      archive_close(info->ar);
      free(info);
      return 0;
    }
    size = octal_value(hdr + 124, 12);
    flen = strlen(hdr);
    if (size == 0 && flen > 0 && hdr[flen - 1] == '/')
      continue;
    if (ar->nfiles % 100 == 0) {
      int s = ar->nfiles + 101 ;

      if ((info->ptr = realloc(info->ptr, sizeof(long) * s)) == NULL ||
	  (info->fname = realloc(info->fname, sizeof(char*) * s)) == NULL ||
	  (info->size = realloc(info->size, sizeof(long) * s)) == NULL) {
	fprintf(stderr, "No enough memory for info\n");
	archive_close(info->ar);
	free(info);
	return 0;
      }
    }
    info->ptr[ar->nfiles] = archive_tell(info->ar);
    info->fname[ar->nfiles] = strdup(hdr);
    info->size[ar->nfiles] = size;
    ar->nfiles++;
    sizeb = (((size) + (TARBLKSIZ-1)) & ~(TARBLKSIZ-1));
    archive_seek(info->ar, sizeb, SEEK_CUR);
  }

  if (ar->nfiles) {
    ar->info = info;
    ar->format = FORMAT_NAME;
    ar->select = tar_archive_select;
    ar->seek = tar_archive_seek;
    ar->tell = tar_archive_tell;
    ar->read = tar_archive_read;
    ar->close = tar_archive_close;
    return 1;
  }
  archive_close(info->ar);
  free(info);
  return 0;
}

int
tar_archive_select(Archive *ar, int index)
{
  TAR_info *info = ar->info;

  if (index < 0)
    return 0;
  archive_seek(info->ar, info->ptr[index], SEEK_SET);

  strncpy(ar->ifname, info->fname[index], 16);
  ar->ifname[15] = '\0';
#ifdef DEBUG
  fprintf(stderr, "name = %s\n", ar->ifname);
#endif

  ar->size = info->size[index];
  ar->offset = info->ptr[index];
  info->pos = 0;
  return 1;
}

int
tar_archive_seek(Archive *ar, long pos, int whence)
{
  TAR_info *info = ar->info;

  switch (whence) {
  case SEEK_SET:
    info->pos = pos;
    return archive_seek(info->ar, ar->offset + pos, SEEK_SET);
  case SEEK_CUR:
    info->pos += pos;
    return archive_seek(info->ar, pos, SEEK_CUR);
  case SEEK_END:
    info->pos = ar->size - 1 + pos;
    return archive_seek(info->ar, ar->offset + ar->size - 1 + pos, SEEK_SET);
  default:
    fprintf(stderr, "tar_archive_seek: FATAL: Invalid whence specified: %d\n", whence);
    exit(-1);
  }
}

int
tar_archive_tell(Archive *ar)
{
  TAR_info *info = ar->info;

  return archive_tell(info->ar) - ar->offset;
}

int
tar_archive_read(Archive *ar, unsigned char *buf, int size)
{
  TAR_info *info = ar->info;
  int r;

  r = (info->pos + size <= ar->size) ?
    archive_read(info->ar, buf, size) :
    archive_read(info->ar, buf, ar->size - info->pos);
  info->pos += r;

  return r;
}

int
tar_archive_close(Archive *ar)
{
  TAR_info *info = ar->info;
  int i, f = archive_close(info->ar);

  for (i = 0; i < ar->nfiles; i++)
    free(info->fname[i]);
  free(info->fname);
  free(info->ptr);
  free(info->size);
  free(ar->info);

  return f;
}
