/*
 * lfg.c -- lfg file handler
 * (C)Copyright 1999 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Tue Sep 21 07:52:12 1999.
 * $Id: lfg.c,v 1.9 1999/09/20 23:48:23 sian Exp $
 *
 * by courtesy of TF
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "plugin.h"
#include "utils.h"
#include "lfg.h"

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
loader_lfg_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Loader;
  p->pluginname = "LFG(LF2) Format Loader Plugin version 0.2.1";
  p->pluginshortname = FORMAT_NAME;
  p->author = "Hiroshi Takekawa";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.loader.load_image = lfg_load_image;

  return 1;
}

static int lfg_read_image(Info *info, Image *p, int direction, int datasize)
{
  Archive *ar = info->ar;
  unsigned char ring[RING_BUFFER_SIZE], *data, flag = 0, upper, lower;
  int i, j, c, pos, len, m, x, y;

  if ((data = (unsigned char *)malloc(datasize)) == NULL) {
    fprintf(stderr, "lfg_read_image: No enough memory for data\n");
    return 0;
  }

  memset(ring, 0, RING_BUFFER_SIZE);
  for (i = 0, c = 0, m = 0xfee; i < datasize;) {
    if (--c < 0) {
      flag = archive_getc(ar);
      c = 7;
    }

    if (flag & 0x80) {
      /* data */
      ring[m] = data[i++] = archive_getc(ar);
      m = (++m) % RING_BUFFER_SIZE;
    } else {
      /* copy */
      upper = archive_getc(ar);
      lower = archive_getc(ar);

      len = (upper & 0xf) + 3;
      pos = (upper >> 4) + (lower << 4);
      for (j = 0; j < len; j++) {
	ring[m] = data[i++] = ring[pos];
	m = (++m) % RING_BUFFER_SIZE;
	pos = (++pos) % RING_BUFFER_SIZE;
      }
    }
    flag <<= 1;
  }

  if (i != datasize) {
    fprintf(stderr, "Extracted size(%d) != data size %d\n", i, datasize);
    free(data);
    return 0;
  }

  if ((p->image = malloc(p->image_size)) == NULL) {
    fprintf(stderr, "No enough memory for image\n");
    free(data);
    return 0;
  }

  x = y = 0;
  for (i = 0; i < datasize; i++) {
    p->image[p->width * y + x    ] =
      (data[i] & 0x80) >> 4 |
      (data[i] & 0x20) >> 3 |
      (data[i] & 0x08) >> 2 |
      (data[i] & 0x02) >> 1;
    p->image[p->width * y + x + 1] =
      (data[i] & 0x40) >> 3 |
      (data[i] & 0x10) >> 2 |
      (data[i] & 0x04) >> 1 |
      (data[i] & 0x01);
    if (direction == VERTICAL) {
      if (++y >= p->height) {
	y = 0;
	x += 2;
      }
    } else {
      x += 2;
      if (x >= p->width) {
	x = 0;
	y++;
      }
    }
  }

  free(data);

  return 1;
}

static int lfg_decode_image(Info *info, Image *p)
{
  Archive *ar = info->ar;
  int i, j, k, u, l, direction, datasize;
  unsigned char c;
  char buf[8];

  if (archive_read(ar, buf, 8) != 8)
    return 0;

  if (memcmp("LEAFCODE", buf, 8))
    return 0;

  info->format = FORMAT_NAME;

#ifdef DEBUG
  fprintf(stderr, "lfg_decode_image: lfg identified\n");
#endif

  /* read palette */
  j = k = 0;
  for (i = 0; i < 24; i++) {
    c = archive_getc(ar);
    u = (c >> 4) & 0xf;
    l = c & 0xf;

    p->colormap[j][k++] = (u << 4) | u;
    k %= 3;
    j += k == 0 ? 1 : 0;
    p->colormap[j][k++] = (l << 4) | l;
    k %= 3;
    j += k == 0 ? 1 : 0;
  }
  p->ncolors = 16;

  /* read location and size */
  p->left = read_little_word(ar) << 3;
  p->top = read_little_word(ar);
  p->width = ((read_little_word(ar) + 1) << 3) - p->left;
  p->height = (read_little_word(ar) + 1) - p->top;

  /* set other variables */
  p->type = _INDEX;
  p->bytes_per_line = p->width;
  p->image_size = p->bytes_per_line * p->height;

  /* read orientation */
  direction = archive_getc(ar) == 0 ? VERTICAL : HORIZONTAL;

  /* read transparent */
  if ((c = archive_getc(ar)) != 0xff) {
    if (c > 15)
      fprintf(stderr, "Invalid transparent color %d\n", (int)c);
    else {
      p->transparent_disposal = info->transparent_disposal;
      p->transparent.index = c;
    }
  }

  /* skip padding (2bytes) */
  i = read_little_word(ar);

  /* read data size */
  datasize = read_little_dword(ar);

  return lfg_read_image(info, p, direction, datasize);
}

int lfg_load_image(Info *info, Image *p)
{
  if (!lfg_decode_image(info, p)) {
    if (info->format != NULL)
      return PROC_ERROR;
    archive_seek(info->ar, 0, SEEK_SET);
    if (!lf2_decode_image(info, p))
      return info->format == NULL ? PROC_NOT : PROC_ERROR;
  }

  info->npics = 1;

  /* set handler */
  info->handler.destroy  = NULL;
  info->handler.alarmset = NULL;
  info->handler.alarm    = NULL;
  info->handler.alarmoff = NULL;

  return PROC_OK;
}
