/*
 * mag.c -- enfle mag handler
 * (C)Copyright 1998, 99 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Tue Sep 21 07:53:10 1999.
 * $Id: mag.c,v 1.10 1999/09/20 23:48:18 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "enfle.h"
#include "mag.h"
#include "utils.h"
#include "plugin.h"

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
loader_mag_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Loader;
  p->pluginname = "MAG Format Loader Plugin version 0.2";
  p->pluginshortname = FORMAT_NAME;
  p->author = "Hiroshi Takekawa";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.loader.load_image = mag_load_image;

  return 1;
}

/* read mag header */
int mag_read_header(Info *info, Image *p, MAG_info *m_info)
{
  Archive *ar = info->ar;
  char buf[22];
  int c, i;

  /* compare check data */
  archive_read(ar, buf, 8);
  if (memcmp(buf, "MAKI02  ", 8))
    return 0;
  info->format = FORMAT_NAME;
  /* read machine code and username */
  archive_read(ar, buf, 4 + 18);

  /* skip memo */
  do {
    c = archive_getc(ar);
  } while (c != 0);

  m_info->offset = archive_tell(ar) - 1;
  m_info->mcode = archive_getc(ar);
  m_info->dcode = archive_getc(ar);
  m_info->screen = archive_getc(ar);
  m_info->lx = read_little_word(ar);
  m_info->uy = read_little_word(ar);
  m_info->rx = read_little_word(ar);
  m_info->dy = read_little_word(ar);
  m_info->off_a = read_little_dword(ar);
  m_info->off_b = read_little_dword(ar);
  m_info->size_b = read_little_dword(ar);
  m_info->off_p = read_little_dword(ar);
  m_info->size_p = read_little_dword(ar);

  p->ncolors = (m_info->screen & COLOR256) ? 256 : 16;
  p->type = _INDEX;

  m_info->lpad =      m_info->lx & 7;
  m_info->rpad = 7 - (m_info->rx & 7);
  m_info->lx -= m_info->lpad;
  m_info->rx += m_info->rpad;

  p->width  = m_info->rx - m_info->lx + 1;
  p->height = m_info->dy - m_info->uy + 1;
  p->left   = m_info->lx;
  p->top    = m_info->uy;

  /* read palette */
  for (i = 0; i < p->ncolors; i++) {
    /* ORDER: GRB */
    p->colormap[i][1] = archive_getc(ar);
    p->colormap[i][0] = archive_getc(ar);
    p->colormap[i][2] = archive_getc(ar);
  }

  m_info->flagperline = p->width >> ((m_info->screen & COLOR256) ? 2 : 3);

  return 1;
}

int mag_decode_image(Info *info, Image *p)
{
  Archive *ar = info->ar;
  int i, j, pr, f, t, sa, width, height;
  byte *flaga, *flagb, *fa, *fb, *pix, *d;
  MAG_info m_info;
  unsigned char bitmask[] = {
    128, 64, 32, 16, 8, 4, 2, 1
  };
  int offset[][2] = {
    { 0,  0},    {-1,  0},    {-2,  0},    {-4,  0},
    { 0, -1},    {-1, -1},    { 0, -2},    {-1, -2},
    {-2, -2},    { 0, -4},    {-1, -4},    {-2, -4},
    { 0, -8},    {-1, -8},    {-2, -8},    { 0, -16}
  };

  if (!mag_read_header(info, p, &m_info))
    return 0;
  width = p->width;
  height = p->height;

  /* memory allocate for flag */
  sa = m_info.off_b - m_info.off_a;
  if ((fa = flaga = malloc(sa)) == NULL)
    return 0;
  if ((fb = flagb = malloc(sa << 3)) == NULL) {
    free(flaga);
    return 0;
  }

  /* decode flag */
  archive_seek(ar, m_info.offset + m_info.off_a, 0);
  archive_read(ar, flaga, sa);
  archive_seek(ar, m_info.offset + m_info.off_b, 0);
  for (i = 0; i < sa; i++) {
    f = *fa++;
    for (j = 0; j < 8; j++)
      *fb++ = (f & bitmask[j]) ? archive_getc(ar) : 0;
  }
  free(flaga);

  /* do xor */
  fb = flagb + m_info.flagperline;
  for (i = 0; i < m_info.flagperline * (height - 1); i++, fb++)
    *fb ^= *(fb - m_info.flagperline);

  /* memory allocate for data  */
  p->bytes_per_line = width;
  p->image_size = p->bytes_per_line * height;
  if ((d = p->image = calloc(1, p->image_size)) == NULL) {
    free(flagb);
    return 0;
  }

  /* read pixel */
  archive_seek(ar, m_info.offset + m_info.off_p, 0);
  if ((pix = calloc(1, m_info.size_p)) == NULL) {
    free(flagb);
    return 0;
  }
  if (archive_read(ar, pix, m_info.size_p) != m_info.size_p) {
    free(flagb);
    fprintf(stderr, "Premature MAG file\n");
    return 0;
  }

  /* put pixels */
  pr = 0;
  fb = flagb;
  for (i = 0; i < m_info.flagperline * height; i++) {
    f = (int)*fb++;
    j = f >> 4;
    if (j) {
      if (m_info.screen & COLOR256) {
	*d++ = *(d + offset[j][0] * 2 + offset[j][1] * width);
	*d++ = *(d + offset[j][0] * 2 + offset[j][1] * width);
      } else {
	*d++ = *(d + offset[j][0] * 4 + offset[j][1] * width);
	*d++ = *(d + offset[j][0] * 4 + offset[j][1] * width);
	*d++ = *(d + offset[j][0] * 4 + offset[j][1] * width);
	*d++ = *(d + offset[j][0] * 4 + offset[j][1] * width);
      }
    } else {
      if (m_info.screen & COLOR256) {
	  *d++ = pix[pr++];
	  *d++ = pix[pr++];
      } else {
	t = pix[pr++];
	*d++ = t >> 4;
	*d++ = t & 15;
	t = pix[pr++];
	*d++ = t >> 4;
	*d++ = t & 15;
      }
    }
    j = f & 15;
    if (j) {
      if (m_info.screen & COLOR256) {
	*d++ = *(d + offset[j][0] * 2 + offset[j][1] * width);
	*d++ = *(d + offset[j][0] * 2 + offset[j][1] * width);
      } else {
	*d++ = *(d + offset[j][0] * 4 + offset[j][1] * width);
	*d++ = *(d + offset[j][0] * 4 + offset[j][1] * width);
	*d++ = *(d + offset[j][0] * 4 + offset[j][1] * width);
	*d++ = *(d + offset[j][0] * 4 + offset[j][1] * width);
      }
    } else {
      if (m_info.screen & COLOR256) {
	*d++ = pix[pr++];
	*d++ = pix[pr++];
      } else {
	t = pix[pr++];
	*d++ = t >> 4;
	*d++ = t & 15;
	t = pix[pr++];
	*d++ = t >> 4;
	*d++ = t & 15;
      }
    }
  }
  free(flagb);
  free(pix);

  return 1;
}

int mag_load_image(Info *info, Image *p)
{
  if (!mag_decode_image(info, p))
    return info->format == NULL ? PROC_NOT : PROC_ERROR;

  /* set handler */
  info->handler.destroy  = NULL;
  info->handler.alarmset = NULL;
  info->handler.alarm    = NULL;
  info->handler.alarmoff = NULL;

  info->npics = 1;

  return PROC_OK;
}
