/*
 * pic.c -- enfle pic handler
 * (C)Copyright 1999 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Mon Oct 18 21:18:48 1999.
 * $Id: pic.c,v 1.1 1999/10/18 12:19:59 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "enfle.h"
#include "pic.h"
#include "archive.h"
#include "bitio.h"
#include "utils.h"
#include "plugin.h"

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
loader_pic_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Loader;
  p->pluginname = "PIC Format Loader Plugin version 0.1";
  p->pluginshortname = FORMAT_NAME;
  p->author = "Hiroshi Takekawa";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.loader.load_image = pic_load_image;

  return 1;
}

static int
pic_read_header(Info *info, Image *p, PIC_info *pinfo)
{
  Archive *ar = info->ar;
  char buf[3], *com;
  int c, i, l, r, g, b;

  /* check ID */
  if (archive_read(ar, buf, 3) != 3)
    return 0;
  if (strncmp(buf, "PIC", 3))
    return 0;

  info->format = FORMAT_NAME;

  /* read comment */
  info->comment = NULL;
  l = 0;
  while ((c = archive_getc(ar)) != 0x1a) {
    if (l % 256 == 0) {
      if ((info->comment = realloc(com = info->comment, l + 256)) == NULL)
	return 0;
      if (com)
	free(com);
      return 0;
    }
    info->comment[l++] = c;
  }
  if (l) {
    info->comment[l] = '\0';
    if ((info->comment = realloc(com = info->comment, strlen(info->comment) + 1)) == NULL) {
      free(com);
      return 0;
    }
  }

  /* skip real comment */
  while (archive_getc(ar)) ;

  /* reserved */
  (void)archive_getc(ar);

  /* type and mode */
  c = archive_getc(ar);
  pinfo->type = c & 15;
  pinfo->mode = c >> 4;
  if (pinfo->type) {
    fprintf(stderr, "pic_decode_image: unsupported type %d\n", pinfo->type);
    return 0;
  }

  pinfo->color_bits = read_big_word(ar);
  p->ncolors = 1 << pinfo->color_bits;
  switch (pinfo->color_bits) {
  case 4:
  case 8:
    pinfo->if_palette = 1;
    pinfo->if_cache = 0;
    p->type = _INDEX;
    break;
  case 15:
  case 16:
    pinfo->if_palette = 0;
    pinfo->if_cache = 1;
    p->type = _RGB24;
    break;
  default:
    fprintf(stderr, "pic_decode_image: unsupported depth %d\n", pinfo->color_bits);
    return 0;
  }

  p->width = read_big_word(ar);
  p->height = read_big_word(ar);

  if (pinfo->if_palette) {
    for (i = 0; i < p->ncolors; i++) {
      c = read_big_word(ar);
      r = (c & 0x07c0) >> 3;
      g = (c & 0xf800) >> 8;
      b = (c & 0x003e) << 2;
      if (c & 1) {
	r |= 7;
        g |= 7;
	b |= 7;
      }
      p->colormap[i][0] = r;
      p->colormap[i][1] = g;
      p->colormap[i][2] = b;
    }
  }

  return 1;
}

static int
pic_decode_image(Info *info, Image *p)
{
  int col, index, len, bits, r, g, b;
  unsigned char *ptr, *flag;
  Dlist *d;
  Dlist_data *dd;
  CColor c[128], color, *cp;
  Bit_stream *st;
  PIC_info pinfo;

#ifdef DEBUG
  fprintf(stderr, "pic_decode_image() called\n");
#endif
  if (!pic_read_header(info, p, &pinfo))
    return 0;
#ifdef DEBUG
  fprintf(stderr, "PIC: %dx%d %d colors\n", p->width, p->height, p->ncolors);
#endif
  if ((d = dlist_create()) == NULL)
    return 0;

  p->bytes_per_line = image_calculate_bytes_per_line(p->width, p->type);
  p->image_size = p->bytes_per_line * p->height;
  if ((p->image = ptr = malloc(p->image_size)) == NULL)
    return 0;
  if ((flag = calloc(1, p->width * p->height)) == NULL) {
    free(p->image);
    p->image = NULL;
    return 0;
  }

  st = bitio_init(info->ar);
  color.r = color.g = color.b = color.i = 0;
  memset(c, 0, sizeof(CColor) * 128);
  c[0].d = dlist_add(d, &c[0]);
  for (index = -1; index < p->width * p->height;) {
    /* read length */
    for (bits = 1; bitio_getbit(st); bits++) ;
    len = (1 << bits) - 1 + bitio_getbits(st, bits);
    /* fill with the current color */
    for (; len > 0; len--, index++)
      if (pinfo.if_palette) {
	if (index >= 0) {
	  if (flag[index])
	    color.i = ptr[index];
	  else
	    ptr[index] = color.i;
	}
      } else {
	if (index >= 0) {
	  if (flag[index]) {
	    color.r = ptr[index * 3 + 0];
	    color.g = ptr[index * 3 + 1];
	    color.b = ptr[index * 3 + 2];
	  } else {
	    ptr[index * 3 + 0] = color.r;
	    ptr[index * 3 + 1] = color.g;
	    ptr[index * 3 + 2] = color.b;
	  }
	}
      }

    if (index == p->width * p->height)
      break;

    /* read new color */
    if (pinfo.if_cache) {
      if (bitio_getbit(st)) {
	/* cached */
	col = bitio_getbits(st, 7);
	color.r = c[col].r;
	color.g = c[col].g;
	color.b = c[col].b;
	dlist_move_to_top(d, c[col].d);
      } else {
	if (d->ndata < 128) {
	  col = d->ndata;
	  g = bitio_getbits(st, 5) << 3;
	  r = bitio_getbits(st, 5) << 3;
	  b = bitio_getbits(st, 5) << 3;
	  if (pinfo.color_bits == 16)
	    if (bitio_getbit(st)) {
	      r |= 7;
	      g |= 7;
	      b |= 7;
	    }
	  color.r = c[col].r = r;
	  color.g = c[col].g = g;
	  color.b = c[col].b = b;
	  c[col].d = dlist_add(d, &c[col]);
	  dlist_move_to_top(d, c[col].d);
	} else {
	  dd = dlist_gethead(d);
	  cp = dd->data;
	  g = bitio_getbits(st, 5) << 3;
	  r = bitio_getbits(st, 5) << 3;
	  b = bitio_getbits(st, 5) << 3;
	  if (pinfo.color_bits == 16)
	    if (bitio_getbit(st)) {
	      r |= 7;
	      g |= 7;
	      b |= 7;
	    }
	  color.r = cp->r = r;
	  color.g = cp->g = g;
	  color.b = cp->b = b;
	  dlist_move_to_top(d, dd);
	}
      }
    } else
      color.i = bitio_getbits(st, pinfo.color_bits);

    /* expand chain */
    if (bitio_getbit(st)) {
      int x = index % p->width;
      int y = index / p->width;
      int f = 0;

      for (;;) {
	y++;
	switch (bitio_getbits(st, 2)) {
	case 1:
	  x--;
	  break;
	case 2:
	  break;
	case 3:
	  x++;
	  break;
	case 0:
	  if (bitio_getbit(st)) {
	    if (bitio_getbit(st))
	      x += 2;
	    else
	      x -= 2;
	  } else
	    f = 1;
	  break;
	default:
	  break;
	}
	if (f)
	  break;
	flag[y * p->width + x] = 1;
	if (pinfo.if_palette)
	  ptr[y * p->width + x] = color.i;
	else {
	  ptr[(y * p->width + x) * 3 + 0] = color.r;
	  ptr[(y * p->width + x) * 3 + 1] = color.g;
	  ptr[(y * p->width + x) * 3 + 2] = color.b;
	}
      }
    }
  }
  free(flag);
  dlist_destroy(d, 0);

#ifdef DEBUG
  fprintf(stderr, "done\n");
#endif
  return 1;
}

int
pic_load_image(Info *info, Image *p)
{
  if (!pic_decode_image(info, p))
    return info->format == NULL ? PROC_NOT : PROC_ERROR;

  info->npics = 1;

  /* set handler */
  info->handler.destroy  = NULL;
  info->handler.alarmset = NULL;
  info->handler.alarm    = NULL;
  info->handler.alarmoff = NULL;

  return PROC_OK;
}
