/*
 * xpm.c -- enfle xpm handler
 * (C)Copyright 1999 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Fri May 19 22:11:46 2000.
 * $Id: xpm.c,v 1.11 2000/05/19 16:46:48 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "enfle.h"
#include "_xpm.h"
#include "hash.h"
#include "parse.h"
#include "plugin.h"
#include "utils.h"

#include <ctype.h>

#define TRANSPARENT_COLOR "None"
#define RGBFILE "/usr/X11R6/lib/X11/rgb.txt"

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
loader_xpm_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Loader;
  p->pluginname = "XPM Format Loader Plugin version 0.2.3";
  p->pluginshortname = FORMAT_NAME;
  p->author = "Hiroshi Takekawa";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.loader.load_image = xpm_load_image;

  return 1;
}

#define GUESS_LENGTH 256
static char *
get_string(Archive *ar)
{
  char *p, *buf;
  int c, count = 0;

  /* skip spaces */
  for (;;) {
    while (isspace(c = archive_getc(ar))) ;

    /* skip comment */
    if (c == '/') {
      if ((c = archive_getc(ar)) != '*') {
	fprintf(stderr, "got %c after /\n", c);
	return NULL;
      }
      for (;;) {
	while (archive_getc(ar) != '*') ;
	if (archive_getc(ar) == '/')
	  break;
      }
    } else
      break;
  }

  /* does not start with a double quotation */
  if (c != '"')
    return NULL;

  /* allocate memory */
  if ((buf = malloc(GUESS_LENGTH)) == NULL)
    return NULL;

  for (;;) {
    /* read string up to GUESS_LENGTH bytes */
    if (archive_gets(ar, buf + count * (GUESS_LENGTH - 1), GUESS_LENGTH) == NULL) {
      free(buf);
      return NULL;
    }

    /* search read string for a double quotation which closes a string */
    for (p = buf + count * (GUESS_LENGTH - 1); *p != '\0' && *p != '"'; p++) ;
    if (*p == '"')
      /* yes, found a double quotation */
      break;
    /* no, more data required */
    count++;
    if ((p = realloc(buf, (count + 1) * GUESS_LENGTH)) == NULL) {
      free(buf);
      return NULL;
    }
    buf = p;
  }

  /* terminate string */
  *p = '\0';

  /* seek archive pointer to assure that we will get next to '"' at next reading. */
  if ((c = buf + strlen(buf) - p + 2))
    archive_seek(ar, -c, SEEK_CUR);

  /* shrink memory */
  if ((p = realloc(buf, strlen(buf) + 1)) == NULL) {
    free(buf);
    return NULL;
  }

  return p;
}

static int
hexchar_to_int(char c)
{
  if (isxdigit(c))
    return (isdigit(c)) ? c - '0' : toupper(c) - 'A' + 10;
  else
    return -1;
}

static int
hextwochars_to_int(char *p)
{
  int c1, c2;

  c1 = hexchar_to_int(p[0]);
  if (c1 < 0)
    return -1;
  c2 = hexchar_to_int(p[1]);
  if (c2 < 0)
    return -1;
  return (c1 << 4) + c2;
}

static int
hexfourchars_to_int(char *p)
{
  int c1, c2;

  c1 = hextwochars_to_int(p);
  if (c1 < 0)
    return -1;
  c2 = hextwochars_to_int(p + 2);
  if (c2 < 0)
    return -1;
  return (c1 << 8) + c2;
}

static char *
shrink_spaces(char *str)
{
  char *new, *p;

  if ((new = malloc(strlen(str) + 1)) == NULL)
    return NULL;
  while (isspace(*str))
    str++;
  for (p = new; *str != '\0';)
    if (isspace(*str)) {
      *p++ = ' ';
      while (isspace(*str))
	str++;
    } else
      *p++ = *str++;
  *p = '\0';
  if ((p = realloc(new, strlen(new) + 1)) == NULL)
    return NULL;
  return p;
}

static int
parse_header(Archive *ar, Image *p, int *cpp)
{
  int i;
  char *ptr, *line, **tokens;

  if ((line = get_string(ar)) == NULL)
    return PROC_ERROR;
  ptr = shrink_spaces(line);
  free(line);
  line = ptr;

  if ((tokens = split(line, ' ')) == NULL)
    return PROC_ERROR;
  free(line);

  /* count a number of arguments */
  for (i = 0; tokens[i] != NULL; i++) ;

  /* four arguments are necessary */
  if (i < 4 || i == 5 || i > 7) {
    free_string_array(tokens);
    return PROC_NOT;
  }
  p->width = atoi(tokens[0]);
  p->height = atoi(tokens[1]);
  p->ncolors = atoi(tokens[2]);
  *cpp = atoi(tokens[3]);

  /* optional hotspot */
  if (i > 4) {
    p->left = atoi(tokens[4]);
    p->top = atoi(tokens[5]);
  }

  /* check extension */
  if (i == 7) {
    if (strcmp(tokens[6], "XPMEXT")) {
      free_string_array(tokens);
      return PROC_NOT;
    }
    printf("XPMEXT found, but ignored\n");
  }

  free_string_array(tokens);

  return PROC_OK;
}

static Hash *
parse_color(Info *info, Image *p, Archive *ar, int cpp, Hash *h)
{
  Hash *colortable;
  char *chars, *line, *ptr, **tokens;
  int c, i, j, n, num, size, t;
  NColor *color, *lcolor;

  /* memory allocate for key */
  if ((chars = malloc(cpp + 1)) == NULL)
    return NULL;
  chars[cpp] = '\0';

  /* create hash for color lookup */
  size = 2017;
  t = p->ncolors >> 8;
  while (t > 1) {
    size <<= 1;
    t >>= 1;
  }
  if ((colortable = hash_create(size)) == NULL)
    return NULL;

  for (n = 0; n < p->ncolors; n++) {
    while (isspace(c = archive_getc(ar))) ;
    if (c != ',') {
      fprintf(stderr, "got %c\n", c);
      hash_destroy(colortable, 1);
      free(chars);
      return NULL;
    }
    if ((line = get_string(ar)) == NULL) {
      hash_destroy(colortable, 1);
      free(chars);
      return NULL;
    }

    /* get chars */
    for (i = 0; i < cpp; i++)
      chars[i] = line[i];

    /* replace multiple '\t', ' ' with single ' '. */
    ptr = shrink_spaces(line + cpp);
    free(line);
    line = ptr;

    /* split */
    if ((tokens = split(line, ' ')) == NULL) {
      hash_destroy(colortable, 1);
      free(chars);
      return NULL;
    }
    free(line);

    /* count a number of arguments */
    for (num = 0; tokens[num] != NULL; num++) ;

    /* parse arguments */
    for (i = 0; i < num;) {
      switch (tokens[i][0]) {
      case 'm':
	i += 2;
	break;
      case 'g':
	if ((tokens[i][1] == '4' && tokens[i][2] == '\0') || tokens[i][1] == '\0') {
	  i += 2;
	  break;
	}
	hash_destroy(colortable, 1);
	free(chars);
	free_string_array(tokens);
	return NULL;
      case 's':
#if 0
	if ((color = hash_lookup(h, tokens[i + 1])) == NULL) {
	    hash_destroy(colortable, 1);
	    free(chars);
	    free_string_array(tokens);
	    return NULL;
	}
	hash_register(colortable, chars, color);
#endif
	i += 2;
	break;
      case 'c':
	i++;
	if (!strcasecmp(tokens[i], TRANSPARENT_COLOR)) {
	  if ((color = malloc(sizeof(NColor))) == NULL) {
	    hash_destroy(colortable, 1);
	    free(chars);
	    free_string_array(tokens);
	    return NULL;
	  }
	  color->c[0] = color->c[1] = color->c[2] = -1;
	  color->n = n;
	  p->transparent.index = n;
	  p->transparent.red = 0;
	  p->transparent.green = 0;
	  p->transparent.blue = 1;
	  p->transparent_disposal = info->transparent_disposal;
	} else if (tokens[i][0] != '#') {
	  /* color names are case insensitive */
	  for (j = 0; j < strlen(tokens[i]); j++)
	    if (isupper(tokens[i][j]))
	      tokens[i][j] = tolower(tokens[i][j]);
	  if ((lcolor = hash_lookup(h, tokens[i])) == NULL) {
	    fprintf(stderr, "color %s not found\n", tokens[i]);
	    hash_destroy(colortable, 1);
	    free(chars);
	    free_string_array(tokens);
	    return NULL;
	  }
	  if ((color = malloc(sizeof(NColor))) == NULL) {
	    hash_destroy(colortable, 1);
	    free(chars);
	    free_string_array(tokens);
	    return NULL;
	  }
	  memcpy(color, lcolor, sizeof(NColor));
	} else {
	  if ((color = malloc(sizeof(NColor))) == NULL) {
	    hash_destroy(colortable, 1);
	    free(chars);
	    free_string_array(tokens);
	    return NULL;
	  }

	  switch (strlen(tokens[i] + 1)) {
	  case 6:
	    /* normal #FFFFFF style */
	    color->c[0] = hextwochars_to_int(tokens[i] + 1);
	    color->c[1] = hextwochars_to_int(tokens[i] + 3);
	    color->c[2] = hextwochars_to_int(tokens[i] + 5);
	    break;
	  case 12:
	    /* 16bits #FFFFFFFFFFFF style */
	    color->c[0] = hexfourchars_to_int(tokens[i] + 1) >> 8;
	    color->c[1] = hexfourchars_to_int(tokens[i] + 5) >> 8;
	    color->c[2] = hexfourchars_to_int(tokens[i] + 9) >> 8;
	    break;
	  case 3:
	    /* 4bits #FFF style */
	    color->c[0] = hexchar_to_int(tokens[i][1]) << 4;
	    color->c[1] = hexchar_to_int(tokens[i][2]) << 4;
	    color->c[2] = hexchar_to_int(tokens[i][3]) << 4;
	    break;
	  default:
	    fprintf(stderr, "unknown color style %s\n", tokens[i]);
	    hash_destroy(colortable, 1);
	    free(chars);
	    free_string_array(tokens);
	    return NULL;
	  }
	}
	if (n < 256) {
	  p->colormap[n][0] = color->c[0];
	  p->colormap[n][1] = color->c[1];
	  p->colormap[n][2] = color->c[2];
	}
	color->n = n;
	hash_register(colortable, chars, color);
	i++;
	break;
      default:
	hash_destroy(colortable, 1);
	free(chars);
	free_string_array(tokens);
	return NULL;
      }
    }
    free_string_array(tokens);
  }

  free(chars);
  return colortable;
}

static int
parse_body_index(Image *p, Archive *ar, Hash *colortable, int cpp)
{
  unsigned char *d = p->image;
  char *chars, *line, *ptr;
  int c, i, j, linelen;
  NColor *color;

  if ((chars = malloc(cpp + 1)) == NULL)
    return 0;
  chars[cpp] = '\0';
  linelen = p->width * cpp;

  for (i = p->height; i > 0; i--) {
    while (isspace(c = archive_getc(ar))) ;
    if (c != ',') {
      free(chars);
      return 0;
    }
    if ((line = get_string(ar)) == NULL) {
      free(chars);
      return 0;
    }
    if (strlen(line) != linelen)
      return 0;
    for (ptr = line; *ptr != '\0';) {
      for (j = 0; j < cpp; j++)
	chars[j] = *ptr++;
      if ((color = hash_lookup(colortable, chars)) == NULL)
	return 0;
      *d++ = color->n;
    }
    free(line);
  }
  free(chars);

  return 1;
}

static int
parse_body_rgb24(Image *p, Archive *ar, Hash *colortable, int cpp)
{
  unsigned char *d = p->image;
  char *chars, *line, *ptr;
  int c, i, j, linelen;
  NColor *color;

  if ((chars = malloc(cpp + 1)) == NULL)
    return 0;
  chars[cpp] = '\0';
  linelen = p->width * cpp;

  for (i = p->height; i > 0; i--) {
    while (isspace(c = archive_getc(ar))) ;
    if (c != ',') {
      free(chars);
      return 0;
    }
    if ((line = get_string(ar)) == NULL) {
      free(chars);
      return 0;
    }
    if (strlen(line) != linelen) {
      free(chars);
      return 0;
    }
    for (ptr = line; *ptr != '\0';) {
      for (j = 0; j < cpp; j++)
	chars[j] = *ptr++;
      if ((color = hash_lookup(colortable, chars)) == NULL) {
	free(chars);
	return 0;
      }
      if (color->c[0] >= 0) {
	/* avoid transparent color */
	if (color->c[0] == 0 &&
	    color->c[1] == 0 &&
	    color->c[2] == 1)
	  color->c[2] = 2;
	*d++ = color->c[0];
	*d++ = color->c[1];
	*d++ = color->c[2];
      } else {
	/* transparent */
	*d++ = 0;
	*d++ = 0;
	*d++ = 1;
      }
    }
    free(line);
  }
  free(chars);

  return 1;
}

static int
xpm_decode_image(Info *info, Image *p)
{
  static int load_rgb = 0;
  static Hash *h;
  Hash *colortable;
  Archive *ar = info->ar;
  int i, cpp;
  char buf[16];

  /* parse rgb.txt */
  if (!load_rgb) {
    if ((h = rgbparse(RGBFILE)) == NULL) {
      fprintf(stderr, "xpm.c: rgbparse error\n");
      return PROC_ERROR;
    }
    load_rgb++;
  }

  /* see if this file is XPM */
  archive_gets(ar, buf, 16);
#ifdef DEBUG
  fprintf(stderr, "XPM checking: %02X%02X%02X%02X\n", buf[0], buf[1], buf[2], buf[3]);
#endif
  if (strncmp(buf, "/* XPM */", 9))
    return PROC_NOT;

#ifdef DEBUG
  fprintf(stderr, "XPM identified\n");
#endif

  /* identified as XPM */
  info->format = FORMAT_NAME;

  /* skip unused data */
  while (archive_getc(ar) != '{') ;

  /* parse header */
  if ((i = parse_header(ar, p, &cpp)) != PROC_OK)
    return i;
#ifdef DEBUG
  fprintf(stderr, "XPM header: %d %d %d %d\n", p->width, p->height, p->ncolors, cpp);
#endif

  p->bytes_per_line = p->width;
  p->type = p->ncolors > 256 ? _RGB24 : _INDEX;
  p->bytes_per_line = image_calculate_bytes_per_line(p->width, p->type);
  p->image_size = p->height * p->bytes_per_line;
  if ((p->image = malloc(p->image_size)) == NULL)
    return PROC_ERROR;

  /* parse colors */
  if ((colortable = parse_color(info, p, ar, cpp, h)) == NULL) {
    free(p->image);
    p->image = NULL;
    return PROC_ERROR;
  }

#ifdef DEBUG
  fprintf(stderr, "parsing body\n");
#endif
  /* parse body */
  if (p->type == _RGB24 ? !parse_body_rgb24(p, ar, colortable, cpp) :
      !parse_body_index(p, ar, colortable, cpp)) {
    hash_destroy(colortable, 1);
    free(p->image);
    p->image = NULL;
    return PROC_ERROR;
  }

  hash_destroy(colortable, 1);
#ifdef DEBUG
  fprintf(stderr, "XPM file parse done.\n");
#endif

  return PROC_OK;
}

int
xpm_load_image(Info *info, Image *p)
{
  int c;

  if ((c = xpm_decode_image(info, p)) != PROC_OK)
    return c;

  info->npics = 1;

  /* set handler */
  info->handler.destroy  = NULL;
  info->handler.alarmset = NULL;
  info->handler.alarm    = NULL;
  info->handler.alarmoff = NULL;

  return PROC_OK;
}
