/*
 * png.c -- save loaded image as png
 * (C)Copyright 1998, 99 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Sun Jul 30 01:20:38 2000.
 * $Id: png.c,v 1.5 2000/08/01 20:51:13 sian Exp $
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "enfle.h"
#include "version.h"
#include "plugin.h"
#include "_png.h"

#include <png.h>

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
saver_png_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Saver;
  p->pluginname = "PNG Format Saver Plugin version 0.1.2";
  p->pluginshortname = FORMAT_NAME;
  p->author = "Hiroshi Takekawa";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.saver.save_image = png_save_image;

  return 1;
}

/* write a png file */
int
png_save_image(Info *info, Image *p, char *outfile)
{
  FILE *fp = NULL;
  png_structp png_ptr;
  png_infop info_ptr;
  png_text comment[2];
  png_uint_32 k;
  png_bytep *row_pointers;
  char *infile = info->filename;
  int width = p->width;
  int height = p->height;

  if ((fp = fopen(outfile, "wb")) == NULL)
    return 0;

  if ((png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)) == NULL) {
    fclose(fp);
    return 0;
  }

  /* Allocate/initialize the image information data. */
  if ((info_ptr = png_create_info_struct(png_ptr)) == NULL) {
    png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
    fclose(fp);
    return 0;
  }

  /* Set error handling. */
  if (setjmp(png_ptr->jmpbuf)) {
    /* If we get here, we had a problem writing the file */
    png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
    fclose(fp);
    return 0;
  }

  /* set up the output control. */
  png_init_io(png_ptr, fp);

  /* Set the image information here.  Width and height are up to 2^31,
   * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
   * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
   * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
   * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
   * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
   * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
   */
#ifdef DEBUG
  fprintf(stderr, "png_save_image: %dx%d, type %d\n", width, height, p->type);
#endif
  switch (p->type) {
  case _INDEX:
#ifdef DEBUG
    fprintf(stderr, "depth 8 image\n");
#endif
    png_set_IHDR(png_ptr, info_ptr, width, height, 8,
		 PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE,
		 PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
    /* set the palette.  REQUIRED for indexed-color images */
    png_set_PLTE(png_ptr, info_ptr, (png_colorp)p->colormap, p->ncolors);
    break;
  case _RGB16:
    if (!image_expand(p, _RGB24)) {
      fclose(fp);
      png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
      return 0;
    }
  case _RGB24:
    png_set_IHDR(png_ptr, info_ptr, width, height, 8,
		 PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
		 PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
    break;
  default:
    fprintf(stderr, "png_save_image: Unsupported type %d.\n", p->type);
    fclose(fp);
    png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
    return 0;
  }

  png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);

  /* Optionally write comments into the image */
  comment[0].key = "Original filename";
  comment[0].text = infile;
  comment[0].compression = PNG_TEXT_COMPRESSION_NONE;
  comment[1].key = "Software";
  comment[1].text = "Created by Enfle " VERSION " / " COPYRIGHT;
  comment[1].compression = PNG_TEXT_COMPRESSION_zTXt;
  png_set_text(png_ptr, info_ptr, comment, 2);

  /* other optional chunks like cHRM, bKGD, tRNS, tIME, oFFs, pHYs, */
  /* note that if sRGB is present the cHRM chunk must be ignored
   * on read and must be written in accordance with the sRGB profile */

  /* Write the file header information. */
  png_write_info(png_ptr, info_ptr);

  /* Allocate memory */
  if ((row_pointers = png_malloc(png_ptr, sizeof(png_bytep) * height)) == NULL) {
    png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
    fclose(fp);
    return 0;
  }

  switch (p->type) {
  case _INDEX:
    for (k = 0; k < height; k++)
      row_pointers[k] = p->image + k * width;
    break;
  case _RGB24:
    for (k = 0; k < height; k++)
      row_pointers[k] = p->image + 3 * k * width;
    break;
  default:
    fprintf(stderr, "png_save_image: FATAL: internal error: type %d cannot be processed\n", p->type);
    png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
    fclose(fp);
    exit(1);
  }

  /* Write the image */
  png_write_image(png_ptr, row_pointers);
  png_write_end(png_ptr, info_ptr);
  png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
  fclose(fp);

  return 1;
}
